/* Pregunta 3: fork y wait
Se necesita una funcin que indique si un nmero es o no primo utilizando
varios procesadores para hacer clculos parciales en procesos separados,
sincronizando sus resultados en paralelo. Para ello, se le pide que
implemente la siguiente funcin:

typedef unsigned long long int ullint_t;
int primo_par(ullint_t n, unsigned int n_procs);

la cual retorna verdadero o falso segn si el nmero n es o no primo,
respectivamente. Dicho clculo debe hacerse usando n_procs procesos en
paralelo. Por motivos de correctitud, una vez finalizada la funcin no debe
haber procesos hijo en estado zombie.

Aquellas soluciones que no usen esta cantidad de procesos o no lo hagan en
paralelo (en forma til) no sern evaluadas. Puede asumir que n_procs es
mayor que cero y que n es lo suficientemente grande como para que valga la
pena usar todos los procesos pedidos. Tambin puede asumir que todas las llamadas a fork() son exitosas y que todos los procesos terminan correctamente.

Hint: Para obtener la informacin de los procesos hijo puede usar la macro
int WEXITSTATUS(int status), que obtiene el cdigo de retorno del hijo a
partir del estado obtenido por wait.
*/

#include <stdlib.h>
#include <unistd.h>
#include <sys/wait.h>

#include "primo_par.h"

static ullint_t min(ullint_t x, ullint_t y) {
	return x <= y ? x : y;
}
static ullint_t max(ullint_t x, ullint_t y) {
	return x >= y ? x : y;
}

int primo_par(ullint_t n, unsigned int n_procs) {
	int es_primo = 1; /* A priori asume que s es primo */
	ullint_t largo_tramo = n / n_procs; /* Tamao del intervalo */

	unsigned int numero_tramo ;
	/* Itera sobre los tramos */
	for (numero_tramo = 0; numero_tramo < n_procs; ++numero_tramo) {
		/* Lmites [inicio, fin) del tramo.
		   Cuidado de no dividir ni por 1 ni por n o ms.
		*/
		ullint_t inicio = max(2, numero_tramo * largo_tramo);
		ullint_t fin = min(n, (numero_tramo + 1) * largo_tramo);
		if (!fork()) {/* Hijo */
			ullint_t factor;
			/* Recorre tramo */
			for (factor = inicio; factor < fin; ++factor) {
				if (n % factor == 0) {/* Si no es primo */
					exit(0);/* Informa al padre */
				}
			}/* Siguiente factor */
			exit(1);/* En el tramo no hay factores */
		}
		/* Crear hijo para siguiente tramo */
	}
	/* Espera a los hijos */
	for (numero_tramo = 0; numero_tramo < n_procs; ++numero_tramo) {
		int status;
		wait(&status);
		/* Si el tramo es primo, el nmero sigue siendo primo.
                   Si el tramo no es primo, el nmero no es primo.
                */
		es_primo = es_primo && WEXITSTATUS(status);
	}
	return es_primo;
}

/* Pauta:

Puntaje:
1.0 Generar n_proc procesos tiles.
0.5 Asignacin correcta de tareas parciales a procesos hijos.
1.0 Calcular bien primalidad parcial en los procesos hijos.
0.5 Informar correctamente al proceso padre.
1.0 Recibir correctamente en el proceso padre el resultado del proceso hijo.
1.0 Calcular bien primalidad global.
1.0 No dejar procesos zombie.

Descuentos:
-1.0 Confusiones padre/hijo por mal uso/entendimiento de fork().
-0.5 Bomba fork finita (grande) o infinita.
-0.5 Tramos de factores no probados.
-6.0 Monoproceso.
-0.5 Prdida de precisin.
-0.5 Tramos exageradamente solapados (repetidos) entre procesos (ms all de
     un par de factores a causa de errores de redondeo).
-0.1 a -0.5 c/u: Errores varios en C.

Bonus:
+0.5 Optimizacin: largo_tramo = ((ullint_t)sqrt(n)) / n_procs;

*/
